searchNodes=[{"doc":"Plug that handles Idempotency-Key HTTP headers. A single Idempotency-Key HTTP header is required for POST and PATCH requests. Handling of requests is based on https://datatracker.ietf.org/doc/draft-ietf-httpapi-idempotency-key-header/ Idempotency Key The value of the Idempotency-Key HTTP header is combined with a URI to produce a unique sha256 hash for the request. This will be used to store the response for first-time requests. The key is used to fetch this response in all subsequent requests. A sha256 hash of the request payload is generated and used to ensure the key is not reused with a different request payload. Error handling By default, errors are raised and handled by the Plug.Exception protocol: Concurrent requests raises IdempotencyPlug.ConcurrentRequestError which sets 409 Conflict HTTP status code. Mismatch of request payload fingerprint will raise IdempotencyPlug.RequestPayloadFingerprintMismatchError which sets 422 Unprocessable Entity HTTP status code. If the first-time request was unexpectedly terminated a IdempotencyPlug.HaltedResponseError which sets a 500 Internal Server error is raised. Setting :with option with an MFA will catch and pass the error to the MFA. Cached responses Cached responses returns an Expires header in the response. See IdempotencyPlug.RequestTracker for more on expiration. Options :tracker - must be a name or PID for the IdempotencyPlug.RequestTracker GenServer, required. :idempotency_key - should be a MFA callback to process idempotency key. Defaults to {Elixir.IdempotencyPlug, :idempotency_key} . :request_payload - should be a MFA to parse request payload. Defaults to {Elixir.IdempotencyPlug, :request_payload} . :hash - should be a MFA to hash an Erlang term. Defaults to {Elixir.IdempotencyPlug, :sha256_hash} . :with - should be one of :exception or MFA. Defaults to :exception . :exception - raises an error. {mod, fun, args} - calls the MFA to process the conn with error, the connection MUST be halted. Examples plug IdempotencyPlug , tracker : IdempotencyPlug.RequestTracker , idempotency_key : { __MODULE__ , :scope_idempotency_key } , request_payload : { __MODULE__ , :limit_request_payload } , hash : { __MODULE__ , :sha512_hash } , with : { __MODULE__ , :handle_error } def scope_idempotency_key ( conn , key ) do { conn . assigns . user . id , key } end def limit_request_payload ( conn ) do Map . drop ( conn . params , [ &quot;value&quot; ] ) end def sha512_hash ( _key , value ) do :sha512 |&gt; :crypto . hash ( :erlang . term_to_binary ( value ) ) |&gt; Base . encode16 ( ) |&gt; String . downcase ( ) end def handle_error ( conn , error ) do conn |&gt; put_status ( error . plug_status ) |&gt; json ( %{ message : error . message } ) |&gt; halt ( ) end","ref":"IdempotencyPlug.html","title":"IdempotencyPlug","type":"module"},{"doc":"Returns the ID as-is.","ref":"IdempotencyPlug.html#idempotency_key/2","title":"IdempotencyPlug.idempotency_key/2","type":"function"},{"doc":"Sorts the request params in a deterministic order.","ref":"IdempotencyPlug.html#request_payload/1","title":"IdempotencyPlug.request_payload/1","type":"function"},{"doc":"Encodes the value from an Erlang term to a binary and generates a sha256 hash from it.","ref":"IdempotencyPlug.html#sha256_hash/2","title":"IdempotencyPlug.sha256_hash/2","type":"function"},{"doc":"A GenServer that tracks processes to ensure requests, at most, are processed once. Storage First-time tracked request will store {:processing, {node, pid}} request state with an expiration date for the provided request ID. Once the request has completed, put_response/3 must be called to update the cached response. The response will be stored as {:ok, data} with an expiration date. The process for a tracked request may halt unexpectedly (e.g. due to raised exception). This module will track the terminated process and store the value as {:halted, reason} . All cached responses will be removed after 24 hours. Lookup For subsequent requests, the state of the first-time tracked request will be returned in the format of {:cache, {:ok, data}, expires_at} . If the request payload fingerprint differs, {:mismatch, {:fingerprint, fingerprint}, expires_at} is returned. If first-time request has not yet completed, {:processing, {node, pid}, expires_at} is returned. If the request unexpectedly terminated, {:cache, {:halted, reason}, expires_at} is returned. Options :cache_ttl - the TTL in milliseconds for any objects in the cache store. Defaults to 24 hours. :prune - the interval in milliseconds to prune the cache store for expired objects. Defaults to 60 seconds. :store - the cache store module to use to store the cache objects. Defaults to {IdempotencyPlug.ETSStore, [table: Elixir.IdempotencyPlug.RequestTracker]} . Examples children = [ { IdempotencyPlug.RequestTracker , cache_ttl : :timer . hours ( 6 ) , prune : :timer . minutes ( 1 ) , store : { IdempotencyPlug.EctoStore , repo : MyApp.Repo } } ] Supervisor . start_link ( children , strategy : :one_for_one , name : MyApp.Supervisor )","ref":"IdempotencyPlug.RequestTracker.html","title":"IdempotencyPlug.RequestTracker","type":"module"},{"doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"IdempotencyPlug.RequestTracker.html#child_spec/1","title":"IdempotencyPlug.RequestTracker.child_spec/1","type":"function"},{"doc":"Updates the state for a given request ID.","ref":"IdempotencyPlug.RequestTracker.html#put_response/3","title":"IdempotencyPlug.RequestTracker.put_response/3","type":"function"},{"doc":"","ref":"IdempotencyPlug.RequestTracker.html#start_link/1","title":"IdempotencyPlug.RequestTracker.start_link/1","type":"function"},{"doc":"Tracks a request ID. This function will return {:init, id, expires_at} for first-time requests. Subsequent requests will return the request state. If the request payload fingerprint differs from what was stored, an error is returned.","ref":"IdempotencyPlug.RequestTracker.html#track/3","title":"IdempotencyPlug.RequestTracker.track/3","type":"function"},{"doc":"Generates a IdempotencyPlug store migration. See Mix.Tasks.Ecto.Gen.Migration for options, takes all options except --change .","ref":"Mix.Tasks.IdempotencyPlug.Ecto.Gen.Migration.html","title":"Mix.Tasks.IdempotencyPlug.Ecto.Gen.Migration","type":"task"},{"doc":"Module that defines an ETS store. The ETS table data is not persisted and will be flushed on application restart. You should use the IdempotencyPlug.EctoStore or a custom IdempotencyPlug.Store type module for production. Examples defmodule MyApp.Application do # .. def start ( _type , _args ) do children = [ { IdempotencyPlug.RequestTracker , [ store : { IdempotencyPlug.EctoStore , [ table : MyAppWeb.RequestTrackerStore ] } ] } # ... ] Supervisor . start_link ( children , strategy : :one_for_one , name : MyApp.Supervisor ) end end","ref":"IdempotencyPlug.ETSStore.html","title":"IdempotencyPlug.ETSStore","type":"module"},{"doc":"Module that defines an Ecto store. A migration file should be generated with mix idempotency_plug.ecto.gen.migration . Examples defmodule MyApp.Application do # .. def start ( _type , _args ) do children = [ { IdempotencyPlug.RequestTracker , [ store : { IdempotencyPlug.EctoStore , repo : MyApp.Repo } ] } # ... ] Supervisor . start_link ( children , strategy : :one_for_one , name : MyApp.Supervisor ) end end","ref":"IdempotencyPlug.EctoStore.html","title":"IdempotencyPlug.EctoStore","type":"module"},{"doc":"Module that defines the store callbacks. Examples defmodule CustomStore do @behaviour IdempotencyPlug.Store @impl true def setup ( options ) @impl true def lookup ( request_id , options ) @impl true def insert ( request_id , fingerprint , data , options ) @impl true def update ( request_id , data , options ) @impl true def prune ( options ) end","ref":"IdempotencyPlug.Store.html","title":"IdempotencyPlug.Store","type":"behaviour"},{"doc":"","ref":"IdempotencyPlug.Store.html#c:insert/5","title":"IdempotencyPlug.Store.insert/5","type":"callback"},{"doc":"","ref":"IdempotencyPlug.Store.html#c:lookup/2","title":"IdempotencyPlug.Store.lookup/2","type":"callback"},{"doc":"","ref":"IdempotencyPlug.Store.html#c:prune/1","title":"IdempotencyPlug.Store.prune/1","type":"callback"},{"doc":"","ref":"IdempotencyPlug.Store.html#c:setup/1","title":"IdempotencyPlug.Store.setup/1","type":"callback"},{"doc":"","ref":"IdempotencyPlug.Store.html#c:update/4","title":"IdempotencyPlug.Store.update/4","type":"callback"},{"doc":"","ref":"IdempotencyPlug.Store.html#t:data/0","title":"IdempotencyPlug.Store.data/0","type":"type"},{"doc":"","ref":"IdempotencyPlug.Store.html#t:expires_at/0","title":"IdempotencyPlug.Store.expires_at/0","type":"type"},{"doc":"","ref":"IdempotencyPlug.Store.html#t:fingerprint/0","title":"IdempotencyPlug.Store.fingerprint/0","type":"type"},{"doc":"","ref":"IdempotencyPlug.Store.html#t:options/0","title":"IdempotencyPlug.Store.options/0","type":"type"},{"doc":"","ref":"IdempotencyPlug.Store.html#t:request_id/0","title":"IdempotencyPlug.Store.request_id/0","type":"type"},{"doc":"There's another request currently being processed for this ID.","ref":"IdempotencyPlug.ConcurrentRequestError.html","title":"IdempotencyPlug.ConcurrentRequestError","type":"exception"},{"doc":"The cached response process didn't terminate correctly.","ref":"IdempotencyPlug.HaltedResponseError.html","title":"IdempotencyPlug.HaltedResponseError","type":"exception"},{"doc":"There are multiple Idempotency-Key request headers.","ref":"IdempotencyPlug.MultipleHeadersError.html","title":"IdempotencyPlug.MultipleHeadersError","type":"exception"},{"doc":"There's no Idempotency-Key request headers.","ref":"IdempotencyPlug.NoHeadersError.html","title":"IdempotencyPlug.NoHeadersError","type":"exception"},{"doc":"The fingerprint for the request payload doesn't match the cached response.","ref":"IdempotencyPlug.RequestPayloadFingerprintMismatchError.html","title":"IdempotencyPlug.RequestPayloadFingerprintMismatchError","type":"exception"},{"doc":"Plug that makes POST and PATCH requests idempotent using Idempotency-Key HTTP header. Follows the IETF Idempotency-Key HTTP Header Field specification draft .","ref":"README.html","title":"IdempotencyPlug","type":"extras"},{"doc":"Add idempotency_plug to your list of dependencies in mix.exs : def deps do [ { :idempotency_plug , &quot;~&gt; 0.2&quot; } ] end","ref":"README.html#installation","title":"IdempotencyPlug - Installation","type":"extras"},{"doc":"First add the request tracker to your supervision tree: defmodule MyApp.Application do # .. def start ( _type , _args ) do children = [ { IdempotencyPlug.RequestTracker , [ name : MyAppWeb.RequestTracker ] } # ... ] Supervisor . start_link ( children , strategy : :one_for_one , name : MyApp.Supervisor ) end end Now add the plug to your pipeline: defmodule MyAppWeb.Router do # ... pipeline :api do plug :accepts , [ &quot;json&quot; ] plug IdempotencyPlug , tracker : MyAppWeb.RequestTracker end # ... end All POST and PATCH requests will now be idempotent using the Idempotency-Key HTTP header value, storing responses with the default ETS store. Persisted store The ETS store is not persisted, so it's not production ready. Instead, let's change the store to use Ecto. First, run mix idempotency_plug.ecto.gen.migration . Now update the configuration for the request tracker: { IdempotencyPlug.RequestTracker , [ store : { IdempotencyPlug.EctoStore , repo : MyApp.Repo } ] } You can also implement your own idempotent request store by using the behaviour in IdempotencyPlug.Store .","ref":"README.html#usage","title":"IdempotencyPlug - Usage","type":"extras"},{"doc":"If you are authenticating users then you must scope the Idempotency-Key to the authenticated user: plug IdempotencyPlug , tracker : MyAppWeb.RequestTracker , idempotency_key : { __MODULE__ , :scope_idempotency_key } def scope_idempotency_key ( conn , id ) , do : { conn . assigns . current_user . id , id } Otherwise you may have a security vulnerability (or conflict) where any user can access another users cached responses when requests are identical.","ref":"README.html#scope-idempotency-key","title":"IdempotencyPlug - Scope Idempotency-Key","type":"extras"},{"doc":"By default, errors are raised and handled by the Plug.Exception protocol, but you can handle the errors setting the :with option: plug IdempotencyPlug , tracker : MyAppWeb.RequestTracker , with : { __MODULE__ , :handle_error } def handle_error ( conn , error ) do conn |&gt; put_status ( Plug.Expetion.Handler . status ( error ) ) |&gt; json ( %{ error : error . message } ) |&gt; halt ( ) end","ref":"README.html#customize-error-response","title":"IdempotencyPlug - Customize error response","type":"extras"},{"doc":"For your controller tests you may want to add this helper to set up the idempotency key: def setup_with_idempotency_key ( %{ conn : conn } ) do conn = Plug.Conn . put_req_header ( conn , &quot;idempotency-key&quot; , Ecto.UUID . bingenerate ( ) ) { :ok , conn : conn } end setup :setup_with_idempotency_key This ensures that all the tests succeed by generating a UUID for all requests.","ref":"README.html#phoenix-tests","title":"IdempotencyPlug - Phoenix tests","type":"extras"},{"doc":"(The MIT License) Copyright (c) 2023 Dan Schultzer &amp; the Contributors Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the 'Software'), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions: The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.","ref":"README.html#license","title":"IdempotencyPlug - LICENSE","type":"extras"},{"doc":"","ref":"CHANGELOG.html","title":"CHANGELOG","type":"extras"},{"doc":"This is a breaking release. If you have been using the Idempotency.Handler behaviour, change your plug to this: plug IdempotencyPlug , tracker : MyAppWeb.RequestTracker , idempotency_key : { MyAppWeb.IdempotencyPlugHandler , :scope_idempotency_key } , with : { MyAppWeb.IdempotencyPlugHandler , :handle_error } And change your handler module to this: defmodule MyAppWeb.IdempotencyPlugHandler do import Phoenix.Controller import Plug.Conn def scope_idempotency_key ( conn , key ) , do : { conn . assigns . current_user . id , key } def handle_error ( conn , error ) do conn |&gt; put_status ( Plug.Exception . status ( error ) ) |&gt; json ( %{ error : error . message } ) |&gt; halt ( ) end end Changes IdempotencyPlug.Handler removed IdempotencyPlug raises errors by default IdempotencyPlug now accepts :idempotency_key , :request_payload , :hash , and :with options IdempotencyPlug now requires :tracker option SHA256 hashing now accepts Erlang terms instead of just binary","ref":"CHANGELOG.html#v0-2-0-2023-04-12","title":"CHANGELOG - v0.2.0 (2023-04-12)","type":"extras"},{"doc":"Fix source url and name in docs","ref":"CHANGELOG.html#v0-1-2-2023-04-07","title":"CHANGELOG - v0.1.2 (2023-04-07)","type":"extras"},{"doc":"Fix indention for generated Ecto migration","ref":"CHANGELOG.html#v0-1-1-2023-04-07","title":"CHANGELOG - v0.1.1 (2023-04-07)","type":"extras"},{"doc":"Initial release","ref":"CHANGELOG.html#v0-1-0-2023-04-06","title":"CHANGELOG - v0.1.0 (2023-04-06)","type":"extras"}]